@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
+# Copyright 2016-2018 Julien Danjou
 # Copyright 2017 Elisey Zanko
 # Copyright 2016 Étienne Bersac
-# Copyright 2016 Julien Danjou
 # Copyright 2016 Joshua Harlow
 # Copyright 2013-2014 Ray Holder
 #
@@ -27,17 +27,14 @@ try:
 except ImportError:
     tornado = None
 
-import inspect
 import sys
 import threading
 from concurrent import futures
 
-from monotonic import monotonic as now
-
 import six
 
 from tenacity import _utils
-from tenacity import wait as _wait
+from tenacity import compat as _compat
 
 # Import all built-in retry strategies for easier usage.
 from .retry import retry_all  # noqa
@@ -49,6 +46,8 @@ from .retry import retry_if_not_result  # noqa
 from .retry import retry_if_result  # noqa
 from .retry import retry_never  # noqa
 from .retry import retry_unless_exception_type  # noqa
+from .retry import retry_if_exception_message  # noqa
+from .retry import retry_if_not_exception_message  # noqa
 
 # Import all nap strategies for easier usage.
 from .nap import sleep  # noqa
@@ -81,6 +80,10 @@ from .before import before_nothing  # noqa
 from .after import after_log  # noqa
 from .after import after_nothing  # noqa
 
+# Import all built-in after strategies for easier usage.
+from .before_sleep import before_sleep_log  # noqa
+from .before_sleep import before_sleep_nothing  # noqa
+
 
 def retry(*dargs, **dkw):
     """Wrap a function with a new `Retrying` object.
@@ -95,7 +98,8 @@ def retry(*dargs, **dkw):
         def wrap(f):
             if asyncio and asyncio.iscoroutinefunction(f):
                 r = AsyncRetrying(*dargs, **dkw)
-            elif tornado and tornado.gen.is_coroutine_function(f):
+            elif tornado and hasattr(tornado.gen, 'is_coroutine_function') \
+                    and tornado.gen.is_coroutine_function(f):
                 r = TornadoRetrying(*dargs, **dkw)
             else:
                 r = Retrying(*dargs, **dkw)
@@ -120,9 +124,72 @@ class DoSleep(float):
     pass
 
 
+class BaseAction(object):
+    """Base class for representing actions to take by retry object.
+
+    Concrete implementations must define:
+    - __init__: to initialize all necessary fields
+    - REPR_ATTRS: class variable specifying attributes to include in repr(self)
+    - NAME: for identification in retry object methods and callbacks
+    """
+
+    REPR_FIELDS = ()
+    NAME = None
+
+    def __repr__(self):
+        state_str = ', '.join('%s=%r' % (field, getattr(self, field))
+                              for field in self.REPR_FIELDS)
+        return '%s(%s)' % (type(self).__name__, state_str)
+
+    def __str__(self):
+        return repr(self)
+
+
+class RetryAction(BaseAction):
+    REPR_FIELDS = ('sleep',)
+    NAME = 'retry'
+
+    def __init__(self, sleep):
+        self.sleep = float(sleep)
+
+
 _unset = object()
 
 
+class RetryError(Exception):
+    """Encapsulates the last attempt instance right before giving up."""
+
+    def __init__(self, last_attempt):
+        self.last_attempt = last_attempt
+        super(RetryError, self).__init__(last_attempt)
+
+    def reraise(self):
+        if self.last_attempt.failed:
+            raise self.last_attempt.result()
+        raise self
+
+    def __str__(self):
+        return "{0}[{1}]".format(self.__class__.__name__, self.last_attempt)
+
+
+class AttemptManager(object):
+    """Manage attempt context."""
+
+    def __init__(self, retry_state):
+        self.retry_state = retry_state
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        if isinstance(exc_value, BaseException):
+            self.retry_state.set_exception((exc_type, exc_value, traceback))
+            return True  # Swallow exception.
+        else:
+            # We don't have the result, actually.
+            self.retry_state.set_result(None)
+
+
 class BaseRetrying(object):
 
     def __init__(self,
@@ -131,23 +198,61 @@ class BaseRetrying(object):
                  retry=retry_if_exception_type(),
                  before=before_nothing,
                  after=after_nothing,
-                 reraise=False):
+                 before_sleep=None,
+                 reraise=False,
+                 retry_error_cls=RetryError,
+                 retry_error_callback=None):
         self.sleep = sleep
-        self.stop = stop
-        self.wait = wait
-        self.retry = retry
-        self.before = before
-        self.after = after
+        self._stop = stop
+        self._wait = wait
+        self._retry = retry
+        self._before = before
+        self._after = after
+        self._before_sleep = before_sleep
         self.reraise = reraise
         self._local = threading.local()
-        # This will allow for passing in the result and handling
-        # the older versions of these functions that do not take
-        # the prior result.
-        self._wait_takes_result = self._waiter_takes_last_result(wait)
+        self.retry_error_cls = retry_error_cls
+        self._retry_error_callback = retry_error_callback
+
+        # This attribute was moved to RetryCallState and is deprecated on
+        # Retrying objects but kept for backward compatibility.
+        self.fn = None
+
+    @_utils.cached_property
+    def stop(self):
+        return _compat.stop_func_accept_retry_state(self._stop)
+
+    @_utils.cached_property
+    def wait(self):
+        return _compat.wait_func_accept_retry_state(self._wait)
+
+    @_utils.cached_property
+    def retry(self):
+        return _compat.retry_func_accept_retry_state(self._retry)
+
+    @_utils.cached_property
+    def before(self):
+        return _compat.before_func_accept_retry_state(self._before)
+
+    @_utils.cached_property
+    def after(self):
+        return _compat.after_func_accept_retry_state(self._after)
+
+    @_utils.cached_property
+    def before_sleep(self):
+        return _compat.before_sleep_func_accept_retry_state(self._before_sleep)
+
+    @_utils.cached_property
+    def retry_error_callback(self):
+        return _compat.retry_error_callback_accept_retry_state(
+            self._retry_error_callback)
 
     def copy(self, sleep=_unset, stop=_unset, wait=_unset,
-             retry=_unset, before=_unset, after=_unset, reraise=_unset):
+             retry=_unset, before=_unset, after=_unset, before_sleep=_unset,
+             reraise=_unset):
         """Copy this object with some parameters changed if needed."""
+        if before_sleep is _unset:
+            before_sleep = self.before_sleep
         return self.__class__(
             sleep=self.sleep if sleep is _unset else sleep,
             stop=self.stop if stop is _unset else stop,
@@ -155,18 +260,10 @@ class BaseRetrying(object):
             retry=self.retry if retry is _unset else retry,
             before=self.before if before is _unset else before,
             after=self.after if after is _unset else after,
+            before_sleep=before_sleep,
             reraise=self.reraise if after is _unset else reraise,
         )
 
-    @staticmethod
-    def _waiter_takes_last_result(waiter):
-        if not six.callable(waiter):
-            return False
-        if isinstance(waiter, _wait.wait_base):
-            waiter = waiter.__call__
-        waiter_spec = inspect.getargspec(waiter)
-        return 'last_result' in waiter_spec.args
-
     def __repr__(self):
         attrs = dict(
             _utils.visible_attrs(self, attrs={'me': id(self)}),
@@ -209,7 +306,7 @@ class BaseRetrying(object):
 
         :param f: A function to wraps for retrying.
         """
-        @six.wraps(f)
+        @_utils.wraps(f)
         def wrapped_f(*args, **kw):
             return self.call(f, *args, **kw)
 
@@ -222,63 +319,65 @@ class BaseRetrying(object):
         return wrapped_f
 
     def begin(self, fn):
-        self.fn = fn
         self.statistics.clear()
-        self.statistics['start_time'] = now()
+        self.statistics['start_time'] = _utils.now()
         self.statistics['attempt_number'] = 1
         self.statistics['idle_for'] = 0
+        self.fn = fn
 
-    def iter(self, result, exc_info, start_time):
-        fut = Future(self.statistics['attempt_number'])
-        if result is not NO_RESULT:
-            trial_end_time = now()
-            fut.set_result(result)
-            retry = self.retry(fut)
-        elif exc_info:
-            trial_end_time = now()
-            t, e, tb = exc_info
-            _utils.capture(fut, exc_info)
-            if isinstance(e, TryAgain):
-                retry = True
-            else:
-                retry = self.retry(fut)
-        else:
+    def iter(self, retry_state):  # noqa
+        fut = retry_state.outcome
+        if fut is None:
             if self.before is not None:
-                self.before(self.fn, self.statistics['attempt_number'])
-
+                self.before(retry_state)
             return DoAttempt()
 
-        if not retry:
+        is_explicit_retry = retry_state.outcome.failed \
+            and isinstance(retry_state.outcome.exception(), TryAgain)
+        if not (is_explicit_retry or self.retry(retry_state=retry_state)):
             return fut.result()
 
         if self.after is not None:
-            trial_time_taken = trial_end_time - start_time
-            self.after(self.fn, self.statistics['attempt_number'],
-                       trial_time_taken)
+            self.after(retry_state=retry_state)
 
-        delay_since_first_attempt = now() - self.statistics['start_time']
         self.statistics['delay_since_first_attempt'] = \
-            delay_since_first_attempt
-        if self.stop(self.statistics['attempt_number'],
-                     delay_since_first_attempt):
+            retry_state.seconds_since_start
+        if self.stop(retry_state=retry_state):
+            if self.retry_error_callback:
+                return self.retry_error_callback(retry_state=retry_state)
+            retry_exc = self.retry_error_cls(fut)
             if self.reraise:
-                raise RetryError(fut).reraise()
-            six.raise_from(RetryError(fut), fut.exception())
+                raise retry_exc.reraise()
+            six.raise_from(retry_exc, fut.exception())
 
         if self.wait:
-            if self._wait_takes_result:
-                sleep = self.wait(self.statistics['attempt_number'],
-                                  delay_since_first_attempt, last_result=fut)
-            else:
-                sleep = self.wait(self.statistics['attempt_number'],
-                                  delay_since_first_attempt)
+            sleep = self.wait(retry_state=retry_state)
         else:
-            sleep = 0
+            sleep = 0.0
+        retry_state.next_action = RetryAction(sleep)
+        retry_state.idle_for += sleep
         self.statistics['idle_for'] += sleep
         self.statistics['attempt_number'] += 1
 
+        if self.before_sleep is not None:
+            self.before_sleep(retry_state=retry_state)
+
         return DoSleep(sleep)
 
+    def __iter__(self):
+        self.begin(None)
+
+        retry_state = RetryCallState(self, fn=None, args=(), kwargs={})
+        while True:
+            do = self.iter(retry_state=retry_state)
+            if isinstance(do, DoAttempt):
+                yield AttemptManager(retry_state=retry_state)
+            elif isinstance(do, DoSleep):
+                retry_state.prepare_for_next_attempt()
+                self.sleep(do)
+            else:
+                break
+
 
 class Retrying(BaseRetrying):
     """Retrying controller."""
@@ -286,23 +385,19 @@ class Retrying(BaseRetrying):
     def call(self, fn, *args, **kwargs):
         self.begin(fn)
 
-        result = NO_RESULT
-        exc_info = None
-        start_time = now()
-
+        retry_state = RetryCallState(
+            retry_object=self, fn=fn, args=args, kwargs=kwargs)
         while True:
-            do = self.iter(result=result, exc_info=exc_info,
-                           start_time=start_time)
+            do = self.iter(retry_state=retry_state)
             if isinstance(do, DoAttempt):
                 try:
                     result = fn(*args, **kwargs)
-                    continue
-                except Exception:
-                    exc_info = sys.exc_info()
-                    continue
+                except BaseException:
+                    retry_state.set_exception(sys.exc_info())
+                else:
+                    retry_state.set_result(result)
             elif isinstance(do, DoSleep):
-                result = NO_RESULT
-                exc_info = None
+                retry_state.prepare_for_next_attempt()
                 self.sleep(do)
             else:
                 return do
@@ -333,23 +428,59 @@ class Future(futures.Future):
         return fut
 
 
-class RetryError(Exception):
-    """Encapsulates the last attempt instance right before giving up."""
+class RetryCallState(object):
+    """State related to a single call wrapped with Retrying."""
 
-    def __init__(self, last_attempt):
-        self.last_attempt = last_attempt
+    def __init__(self, retry_object, fn, args, kwargs):
+        #: Retry call start timestamp
+        self.start_time = _utils.now()
+        #: Retry manager object
+        self.retry_object = retry_object
+        #: Function wrapped by this retry call
+        self.fn = fn
+        #: Arguments of the function wrapped by this retry call
+        self.args = args
+        #: Keyword arguments of the function wrapped by this retry call
+        self.kwargs = kwargs
 
-    def reraise(self):
-        if self.last_attempt.failed:
-            raise self.last_attempt.result()
-        raise self
+        #: The number of the current attempt
+        self.attempt_number = 1
+        #: Last outcome (result or exception) produced by the function
+        self.outcome = None
+        #: Timestamp of the last outcome
+        self.outcome_timestamp = None
+        #: Time spent sleeping in retries
+        self.idle_for = 0
+        #: Next action as decided by the retry manager
+        self.next_action = None
 
-    def __str__(self):
-        return "RetryError[{0}]".format(self.last_attempt)
+    @property
+    def seconds_since_start(self):
+        if self.outcome_timestamp is None:
+            return None
+        return self.outcome_timestamp - self.start_time
+
+    def prepare_for_next_attempt(self):
+        self.outcome = None
+        self.outcome_timestamp = None
+        self.attempt_number += 1
+        self.next_action = None
+
+    def set_result(self, val):
+        ts = _utils.now()
+        fut = Future(self.attempt_number)
+        fut.set_result(val)
+        self.outcome, self.outcome_timestamp = fut, ts
+
+    def set_exception(self, exc_info):
+        ts = _utils.now()
+        fut = Future(self.attempt_number)
+        _utils.capture(fut, exc_info)
+        self.outcome, self.outcome_timestamp = fut, ts
 
 
 if asyncio:
-    from tenacity.async import AsyncRetrying
+    from tenacity._asyncio import AsyncRetrying
 
 if tornado:
     from tenacity.tornadoweb import TornadoRetrying
